package com.caimei365.commodity.service.impl;

import com.aliyun.opensearch.sdk.dependencies.com.google.common.collect.Lists;
import com.aliyun.opensearch.sdk.dependencies.com.google.common.collect.Maps;
import com.aliyun.opensearch.sdk.dependencies.org.json.JSONArray;
import com.aliyun.opensearch.sdk.dependencies.org.json.JSONException;
import com.aliyun.opensearch.sdk.dependencies.org.json.JSONObject;
import com.aliyun.opensearch.sdk.generated.commons.OpenSearchClientException;
import com.aliyun.opensearch.sdk.generated.commons.OpenSearchException;
import com.aliyun.opensearch.sdk.generated.search.*;
import com.caimei365.commodity.components.SearchOpenService;
import com.caimei365.commodity.mapper.SearchMapper;
import com.caimei365.commodity.service.SearchQueryService;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import javax.annotation.Resource;
import java.util.Map;

/**
 * Description
 *
 * @author : Charles
 * @date : 2021/4/7
 */
@Slf4j
@Service
public class SearchQueryServiceImpl implements SearchQueryService {

    private static final Integer SEARCH_MAX_NUM = 500;
    @Value("${caimei.wwwDomain}")
    private String domain;
    @Resource
    private SearchMapper searchMapper;
    @Resource
    private SearchOpenService searchOpenService;

    /**
     * 获取索引Id
     *
     * @param docType 文档类型
     * @param docId   文档ID
     * @return int
     */
    @Override
    public  Integer getIdByDocId(String docType, Integer docId) {
        // 定义Config对象，用于设定config子句参数，指定应用名，分页，数据返回格式等等
        Config config = new Config(Lists.newArrayList(searchOpenService.getAppName()));
        config.setStart(0);
        config.setHits(1);
        // 设置搜索结果返回应用中哪些字段
        config.setFetchFields(Lists.newArrayList("id"));
        // 设置返回格式为json格式
        config.setSearchFormat(SearchFormat.JSON);
        // 创建参数对象
        SearchParams searchParams = new SearchParams(config);
        if (null != docId) {
            if("product".equals(docType)) {
                // 根据商品Id获取索引Id
                searchParams.setQuery("p_id:'" + docId + "'");
            } else if("shop".equals(docType)) {
                // 根据供应商Id获取索引Id
                searchParams.setQuery("s_id:'" + docId + "'");
            } else if("equipment".equals(docType)) {
                // 根据项目仪器Id获取索引Id
                searchParams.setQuery("e_id:'" + docId + "'");
            } else if("article".equals(docType)) {
                // 根据文章Id获取索引Id
                searchParams.setQuery("a_id:'" + docId + "'");
            }
        } else {
            // 创建sort对象，并设置二维排序
            Sort sorter = new Sort();
            // 设置id字段降序
            sorter.addToSortFields(new SortField("id", Order.DECREASE));
            // 若id相同则以RANK相关性算分升序
            sorter.addToSortFields(new SortField("RANK", Order.INCREASE));
            //添加Sort对象参数
            searchParams.setSort(sorter);
        }
        try {
            // 推送查询操作
            String jsonStr = searchOpenService.pushQueryDocument(searchParams);
            // 处理查询结果
            JSONObject jsonObj = new JSONObject(jsonStr);
            String status = jsonObj.getString("status");
            String requestId = jsonObj.getString("request_id");
            String tracer = jsonObj.getString("tracer");
            JSONObject result = jsonObj.getJSONObject("result");
            JSONArray errors = jsonObj.getJSONArray("errors");
            log.info(">>>>>>query id("+docType+"): " + status + ",request_id=" +requestId+ ",errors=" + errors.toString() + ",tracer="+tracer);
            if ("OK".equals(status) && result.getJSONArray("items").length()>0) {
                return result.getJSONArray("items").getJSONObject(0).getInt("id");
            } else {
                return -1;
            }
        } catch (OpenSearchClientException | OpenSearchException | JSONException e) {
            log.error("查询文档索引Id异常：" + e);
            return -1;
        }
    }

    /**
     * 根据文档类型获取索引Id集合
     *
     * @param docType  文档类型
     * @param pageNum   页码
     * @param pageSize  每页数量
     * @return map
     */
    @Override
    public Map<Integer, Integer> getMainIdsByDocType(String docType, int pageNum, int pageSize) {
        String queryStr;
        String idName;
        Map<Integer, Integer> idMap = Maps.newLinkedHashMap();
        if("product".equals(docType)) {
            queryStr = "p_all:'1'";
            idName = "p_id";
        } else if("shop".equals(docType)) {
            queryStr = "s_all:'1'";
            idName = "s_id";
        } else if("equipment".equals(docType)) {
            queryStr = "e_all:'1'";
            idName = "e_id";
        } else if("article".equals(docType)) {
            queryStr = "a_all:'1'";
            idName = "a_id";
        } else {
            return idMap;
        }
        // 定义Config对象，用于设定config子句参数，指定应用名，分页，数据返回格式等等
        Config config = new Config(Lists.newArrayList(searchOpenService.getAppName()));
        config.setStart(pageNum-1);
        config.setHits(pageSize);
        // 设置搜索结果返回应用中哪些字段
        config.setFetchFields(Lists.newArrayList("id", idName));
        // 设置返回格式为json格式
        config.setSearchFormat(SearchFormat.JSON);
        // 创建参数对象
        SearchParams searchParams = new SearchParams(config);
        searchParams.setQuery(queryStr);
        try {
            // 推送查询操作
            String jsonStr = searchOpenService.pushQueryDocument(searchParams);
            // 处理查询结果
            JSONObject jsonObj = new JSONObject(jsonStr);
            String status = jsonObj.getString("status");
            String requestId = jsonObj.getString("request_id");
            String tracer = jsonObj.getString("tracer");
            JSONObject result = jsonObj.getJSONObject("result");
            JSONArray errors = jsonObj.getJSONArray("errors");
            log.info(">>>>>>query idMap("+idName+"): " + status + ",request_id=" +requestId+ ",errors=" + errors.toString() + ",tracer="+tracer);
            if ("OK".equals(status)) {
                JSONArray jsonArray = result.getJSONArray("items");
                for (int i = 0; i < jsonArray.length(); i++) {
                    idMap.put(jsonArray.getJSONObject(i).getInt(idName), jsonArray.getJSONObject(i).getInt("id"));
                }
            }
            return idMap;
        } catch (OpenSearchClientException | OpenSearchException | JSONException e) {
            log.error("查询文档索引Id异常：" + e);
            return idMap;
        }
    }

    /**
     * 根据文档类型获取数量
     *
     * @param docType 文档类型
     * @return int
     */
    @Override
    public Integer getRecordByDocType(String docType) {
        // 定义Config对象，用于设定config子句参数，指定应用名，分页，数据返回格式等等
        Config config = new Config(Lists.newArrayList(searchOpenService.getAppName()));
        config.setStart(0);
        config.setHits(1);
        // 设置搜索结果返回应用中哪些字段
        config.setFetchFields(Lists.newArrayList("id"));
        // 设置返回格式为json格式
        config.setSearchFormat(SearchFormat.JSON);
        // 创建参数对象
        SearchParams searchParams = new SearchParams(config);
        if("product".equals(docType)) {
            searchParams.setQuery("p_all:'1'");
        } else if("shop".equals(docType)) {
            searchParams.setQuery("s_all:'1'");
        } else if("equipment".equals(docType)) {
            searchParams.setQuery("e_all:'1'");
        } else if("article".equals(docType)) {
            searchParams.setQuery("a_all:'1'");
        } else {
            return 0;
        }
        //设置统计子句
        Aggregate agg = new Aggregate();
        agg.setGroupKey(docType);
        agg.setAggFun("count()");
        agg.setAggFilter(docType + "=1");
        //添加Aggregate对象参数
        searchParams.addToAggregates(agg);
        try {
            // 推送查询操作
            String jsonStr = searchOpenService.pushQueryDocument(searchParams);
            // 处理查询结果
            JSONObject jsonObj = new JSONObject(jsonStr);
            String status = jsonObj.getString("status");
            String requestId = jsonObj.getString("request_id");
            String tracer = jsonObj.getString("tracer");
            JSONObject result = jsonObj.getJSONObject("result");
            JSONArray errors = jsonObj.getJSONArray("errors");
            log.info(">>>>>>query  count("+docType+"): " + status + ",request_id=" +requestId+ ",errors=" + errors.toString() + ",tracer="+tracer);
            if ("OK".equals(status)) {
                return result.getJSONArray("facet").getJSONObject(0)
                             .getJSONArray("items").getJSONObject(0).getInt("count");
            } else {
                return -1;
            }
        } catch (OpenSearchClientException | OpenSearchException | JSONException e) {
            log.error("查询文档索引Id异常：" + e);
            return -1;
        }
    }

    /**
     * 根据文档类型获取数量(去重用)
     *
     * @param docType 文档类型
     * @param docId   文档ID
     * @return int
     */
    @Override
    public Integer getCountByDocId(String docType, Integer docId) {
        // 定义Config对象，用于设定config子句参数，指定应用名，分页，数据返回格式等等
        Config config = new Config(Lists.newArrayList(searchOpenService.getAppName()));
        config.setStart(0);
        config.setHits(1);
        // 设置搜索结果返回应用中哪些字段
        config.setFetchFields(Lists.newArrayList("id"));
        // 设置返回格式为json格式
        config.setSearchFormat(SearchFormat.JSON);
        // 创建参数对象
        SearchParams searchParams = new SearchParams(config);
        if (null != docId) {
            if("product".equals(docType)) {
                searchParams.setQuery("p_id:'" + docId + "'");
            } else if("shop".equals(docType)) {
                searchParams.setQuery("s_id:'" + docId + "'");
            } else if("equipment".equals(docType)) {
                searchParams.setQuery("e_id:'" + docId + "'");
            } else if("article".equals(docType)) {
                searchParams.setQuery("a_id:'" + docId + "'");
            } else {
                return 0;
            }
        } else {
            // 创建sort对象，并设置二维排序
            Sort sorter = new Sort();
            // 设置id字段降序
            sorter.addToSortFields(new SortField("id", Order.DECREASE));
            // 若id相同则以RANK相关性算分升序
            sorter.addToSortFields(new SortField("RANK", Order.INCREASE));
            //添加Sort对象参数
            searchParams.setSort(sorter);
        }
        try {
            // 推送查询操作
            String jsonStr = searchOpenService.pushQueryDocument(searchParams);
            // 处理查询结果
            JSONObject jsonObj = new JSONObject(jsonStr);
            String status = jsonObj.getString("status");
            String requestId = jsonObj.getString("request_id");
            String tracer = jsonObj.getString("tracer");
            JSONObject result = jsonObj.getJSONObject("result");
            JSONArray errors = jsonObj.getJSONArray("errors");
            log.info(">>>>>>query count("+docType+"): " + status + ",request_id=" +requestId+ ",errors=" + errors.toString() + ",tracer="+tracer);
            if ("OK".equals(status) && result.getJSONArray("items").length()>=0) {
                return result.getJSONArray("items").length();
            } else {
                return -1;
            }
        } catch (OpenSearchClientException | OpenSearchException | JSONException e) {
            log.error("查询文档索引数量异常：" + e);
            return -1;
        }
    }


}
